# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from os import path
from poort import Request
from poort.utils import FileStorage
from utils import get_environ
from utils import HERE


class TestRequest(object):
    def test_base(self):
        environ = get_environ('base')
        request = Request(environ)

        assert request.as_dict() == {
            'host': 'localhost:8000',
            'path': '/',
            'method': 'GET',
            'ssl': False,

            'ip_address': '127.0.0.1',

            'accept_type': ('text/html,application/xhtml+xml,application/xml;'
                            'q=0.9,image/webp,*/*;q=0.8'),
            'content_type': '',

            'want_json': False,
            'want_html': True,

            'query': {},
            'form': {},
            'files': {},
            'cookies': {},
            'json': None,

            'params': {},
        }

    def test_cookie(self):
        environ = get_environ('cookie-get')
        request = Request(environ)

        assert request.get_cookie('some-cookie') == 'test'

    def test_post(self):
        environ = get_environ('multipart-post')
        request = Request(environ)

        with open(path.join(HERE, 'environs', 'attachment.txt')) as stream:
            attachment = FileStorage(stream, 'attachment.txt', 'attachment',
                                     'text/plain')

            assert request.query == {
                'id': '12',
            }
            assert request.form == {
                'color': ['red', 'green', 'blue'],
                'name': 'poort',
            }
            assert request.files == {
                'attachment': attachment,
            }

            assert request.params == {
                'color': ['red', 'green', 'blue'],
                'name': 'poort',
                'id': '12',
                'attachment': attachment,
            }

    def test_json(self):
        environ = get_environ('json-post')
        request = Request(environ)

        assert request.params == {
            'color': ['red', 'green', 'blue'],
        }

    def test_json_invalid(self):
        environ = get_environ('json-invalid-post')
        request = Request(environ)

        assert request.json is None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from os import path
from poort.utils import environ_from_dict
import yaml

HERE = path.abspath(path.dirname(__file__))


def _load_environ(name):
    with open(path.join(HERE, 'environs', name + '.yaml')) as stream:
        return yaml.safe_load(stream)


def get_environ(name):
    if name == 'base':
        environ = _load_environ('base')
    else:
        environ = _load_environ('base')
        environ.update(_load_environ(name))

    return environ_from_dict(environ)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from poort import Gate
from pytest import raises
from threading import Thread
from time import sleep


class TestGate(object):
    def test_basics(self):
        gate = Gate()
        gate.setup(dict())

        gate.attach('text', 'Hallo world!')
        assert gate.retrieve('text') == 'Hallo world!'
        assert gate.contains('text')
        gate.release('text')
        assert not gate.contains('text')

        gate.teardown()

        with raises(RuntimeError):
            gate.teardown()

    def test_threaded(self):
        gate = Gate()

        def wrap(nr, sleep_time):
            with gate(dict()):
                gate.attach('counter', nr)

                sleep(sleep_time)

                assert gate.contains('counter')
                assert gate.retrieve('counter') == nr

                assert len(gate.local.gate_registered_names)

        thread_1 = Thread(target=wrap, args=(1, 0.2,))
        thread_2 = Thread(target=wrap, args=(2, 0.1,))

        thread_1.start()
        thread_2.start()

        thread_1.join()
        thread_2.join()

        assert not hasattr(gate.local, 'gate_registered_names')

    def test_exceptions(self):
        gate = Gate()
        gate.setup(dict())

        with raises(RuntimeError):
            gate.setup(dict())

        with raises(ValueError):
            gate.retrieve('does-not-exist')

        gate.teardown()

        with raises(RuntimeError):
            gate.teardown()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .request import (
    Request,
)

from .response import (
    FileResponse,
    HtmlResponse,
    JsonResponse,
    Response,
    template_or_json_response,
    TemplateResponse,
    TerminatingResponse,
    wrapped_template_or_json_response,
    WrappedTemplateResponse,
)

from .utils import (
    disinfect_or_json_400,
)

from .gate import (
    Gate,
)

from . import (
    utils,
)

__version__ = '0.3.4'

__package__ = 'poort'
__title__ = 'Poort'
__description__ = 'Poort: The quick gateway.'
__uri__ = 'https://github.com/corverdevelopment/Poort/'

__author__ = 'Nils Corver'
__email__ = 'nils@corverdevelopment.nl'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2016 Corver Development B.V.'

__all__ = [
    'Request',

    'FileResponse',
    'HtmlResponse',
    'JsonResponse',
    'Response',
    'template_or_json_response',
    'TemplateResponse',
    'TerminatingResponse',
    'wrapped_template_or_json_response',
    'WrappedTemplateResponse',

    'Gate',

    'disinfect_or_json_400',

    'utils',
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .request import Request
from threading import local

_local = local()


class Gate(object):
    """Gate is a local storage container (DI of some sorts).

    You should extend this class and provide extra methods and properties
    that you use throughout your application.

    .. code-block:: python

        import poort

        gate = poort.Gate()

        def application(environ, start_response):
            with gate(environ):
                request = gate.request

                if request.path == '/':
                    response = poort.Response('Hallo world!')
                else:
                    response = poort.Response('Not found.', status_code=404)

                return response(request, start_response)

    How to extend `poort.Gate`

    .. code-block:: python

        import poort


        class Gate(poort.Gate):
            def setup(self, environ, *args, **kwargs):
                super(Gate, self).setup(environ, *args, **kwargs)

                dbname = 'test-app'
                self.attach('database', Database(dbname))

            @property
            def database(self):
                ''':rtype: Database'''
                return self.retrieve('database')


        gate = Gate()


        def application(environ, start_response):
            with gate(environ):
                request = gate.request
                database = gate.database

                if request.path == '/':
                    response = poort.Response('Hallo world!')
                else:
                    response = poort.Response('Not found.', status_code=404)

                return response(request, start_response)

    """

    def __init__(self):
        self.local = local()

    def setup(self, environ, *args, **kwargs):
        """Override this method when extending `poort.Gate`.

        :param environ: Environment dict of the WSGI request.

        :type environ: dict

        """
        if hasattr(self.local, 'gate_registered_names'):
            raise RuntimeError('Gate already started.')

        self.local.gate_registered_names = set()
        self.attach('request', Request(environ))

    def teardown(self):
        if not hasattr(self.local, 'gate_registered_names'):
            raise RuntimeError('Gate not started.')

        for name in self.local.gate_registered_names:
            delattr(self.local, name)
        delattr(self.local, 'gate_registered_names')

    def __call__(self, *args, **kwargs):
        self.setup(*args, **kwargs)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.teardown()

    def attach(self, name, value):
        self.local.gate_registered_names.add(name)
        setattr(self.local, name, value)

    def contains(self, name):
        return name in self.local.gate_registered_names

    def retrieve(self, name):
        if not self.contains(name):
            raise ValueError('Property `{:s}` is not attached.'.format(
                name))
        return getattr(self.local, name)

    def release(self, name):
        self.local.gate_registered_names.remove(name)
        delattr(self.local, name)

    @property
    def request(self):
        """

        :rtype: poort.Request
        """
        return self.retrieve('request')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from copy import deepcopy
from io import BytesIO
from pandora.compat import iteritems
from pandora.compat import PY2
from pandora.inflect import underscore
from requests import Request
from requests.packages.urllib3.util import parse_url
from werkzeug.datastructures import FileStorage as WerkzeugFileStorage
from werkzeug.utils import secure_filename
from werkzeug.wsgi import FileWrapper
import re


def flatten_multidict(data, validate=None, wrap=None):
    result = {}
    for key, values in data.lists():
        if validate is not None:
            values = list(filter(validate, values))
        if wrap is not None:
            values = list(map(wrap, values))
        result[key] = values[0] if len(values) == 1 else values
    return result


class FileStorage(WerkzeugFileStorage):
    @classmethod
    def from_original(cls, original):
        return cls(stream=original.stream, filename=original.filename,
                   name=original.name, content_type=original.content_type,
                   content_length=original.content_length,
                   headers=original.headers)

    @property
    def secure_filename(self):
        filename = re.split(r'[\\/]', self.filename)[-1]
        return secure_filename(filename)

    def __eq__(self, other):
        return (self.content_type == other.content_type and
                self.content_length == other.content_length and
                self.mimetype == other.mimetype and
                self.mimetype_params == other.mimetype_params and
                self.filename == other.filename and
                self.secure_filename == other.secure_filename)


def build_environ(method, path, host='localhost',
                  accept_type='text/html', content_type=None,
                  query=None, form=None, files=None, cookies=None):
    if '://' in host:
        url = host.rstrip('/') + '/' + path.lstrip('/')
    else:
        url = 'http://' + host.strip('/') + '/' + path.lstrip('/')

    request = Request(method, url, None, files, form, query, cookies)
    prepared = request.prepare()
    parsed_url = parse_url(prepared.url)

    environ = {
        'HTTP_HOST': parsed_url.host,
        'PATH_INFO': parsed_url.path,
        'REQUEST_METHOD': prepared.method,

        'HTTP_ACCEPT': accept_type,

        'QUERY_STRING': parsed_url.query or '',
    }

    for key, value in iteritems(prepared.headers):
        key = underscore(key)
        if key not in ['content_type', 'content_length']:
            key = 'http_' + key
        environ[key.upper()] = value

    if content_type is not None:
        environ['CONTENT_TYPE'] = content_type

    environ['wsgi.input'] = BytesIO(prepared.body)

    return environ


def mock(app, environ):
    data = list()

    def start_response(status, headers):
        data.append(status)
        data.append(headers)

    data.append(u''.join(app(environ, start_response)))
    return data


def copy_request_environ(environ):
    copy = deepcopy(environ)

    # copy the input and place it back
    copy['INPUT'] = environ['wsgi.input'].read()
    environ['wsgi.input'] = BytesIO(copy['INPUT'])

    del copy['wsgi.errors']
    del copy['wsgi.file_wrapper']
    del copy['wsgi.input']

    if 'gunicorn.socket' in copy:
        del copy['gunicorn.socket']

    return copy


def environ_from_dict(base):
    environ = deepcopy(base)

    environ['wsgi.file_wrapper'] = FileWrapper
    environ['wsgi.errors'] = BytesIO()

    if 'INPUT' in environ:
        if 'CONTENT_LENGTH' not in environ:
            environ['CONTENT_LENGTH'] = str(len(environ['INPUT']))

        if PY2:
            environ['wsgi.input'] = BytesIO(
                bytearray(environ['INPUT'], 'UTF8'))
        else:
            environ['wsgi.input'] = BytesIO(bytes(environ['INPUT'], 'UTF8'))

        del environ['INPUT']
    else:
        environ['wsgi.input'] = BytesIO(b'')

    return environ


def disinfect_or_json_400(params, mapping):
    from .response import JsonResponse
    from .response import TerminatingResponse
    import disinfect as d

    try:
        return mapping(params)
    except d.MappedValueError as exc:
        raise TerminatingResponse(JsonResponse({
            'message': str(exc),
            'errors': exc.errors,
        }, status_code=400))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from io import BytesIO
import poort as p
from utils import get_environ
import disinfect as d
from pytest import raises


class TestRequest(object):
    def test_copy_request_environ(self):
        environ = get_environ('multipart-post')
        assert isinstance(environ['wsgi.input'], BytesIO)

        environ['gunicorn.socket'] = True
        assert 'wsgi.errors' in environ

        copy = p.utils.copy_request_environ(environ)

        assert 'wsgi.errors' not in copy
        assert copy['INPUT'] == environ['wsgi.input'].read()

    def test_without_gunicorn(self):
        environ = get_environ('multipart-post')

        assert 'gunicorn.socket' not in environ

        copy = p.utils.copy_request_environ(environ)

        assert 'wsgi.errors' not in copy
        assert copy['INPUT'] == environ['wsgi.input'].read()

    def test_disinfect_or_json_400(self):
        mapping = d.Mapping({
            'first': d.String(),
        })

        with raises(p.TerminatingResponse) as exc:
            p.disinfect_or_json_400({
                'last': 'Corver'
            }, mapping)

        assert exc.value.response
        assert isinstance(exc.value.response, p.JsonResponse)

        data = exc.value.response.data

        assert exc.value.response.status_code == 400
        assert data['message'] == 'Validation error.'
        assert data['errors']['first'] == 'Field is required.'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from decimal import Decimal
from os import path
from poort import FileResponse
from poort import HtmlResponse
from poort import JsonResponse
from poort import Request
from poort import Response
from poort import template_or_json_response
from poort import TemplateResponse
from poort.utils import FileStorage
from pytest import raises
from utils import get_environ
from utils import HERE
import datetime as dt
import jinja2


class TemplateResponseTest(TemplateResponse):
    renderer = jinja2.Environment(
        loader=jinja2.FileSystemLoader(HERE))


class DummyRequest(object):
    def __init__(self, environ):
        self.environ = environ


class TestRequest(object):
    def test_basics(self):
        data = 'Hello world!'

        response = Response(data)
        response.set_cookie('some-cookie', 'test')
        response.del_cookie('another-cookie')

        assert response.get_status() == '200 OK'
        assert response.get_body() == 'Hello world!'
        assert response.prepare_response() == [b'Hello world!']

        result = list()

        def start_response(status, headers):
            result.append(status)
            result.append(headers)

        result.append(response(None, start_response))

        del_cookie = result[1][1][1]

        assert result == [
            '200 OK',
            [
                ('Set-Cookie', 'some-cookie=test; HttpOnly; Path=/'),
                ('Set-Cookie', del_cookie),
            ],
            [b'Hello world!'],
        ]

        assert 'another-cookie=;' in del_cookie
        assert 'Max-Age=0;' in del_cookie

    def test_json(self):
        class DictClass(object):
            def to_dict(self):
                return {'a': 12}

        class ListClass(object):
            def __iter__(self):
                for i in range(5):
                    yield i

        class StrClass(object):
            def __str__(self):
                return 'A piece of text.'

        today = dt.date.today()
        isoformat = today.isoformat()

        with open(path.join(HERE, 'environs', 'attachment.txt')) as stream:
            attachment = FileStorage(stream, 'attachment.txt', 'attachment',
                                     'text/plain')

        data = {
            'message': 'Hello world!',
            'today': today,
            'undefined': jinja2.Undefined(),
            'decimal': Decimal('1.0008'),
            'filestorage': attachment,
            'dict': DictClass(),
            'list': ListClass(),
            'str': StrClass(),
            'repr': self,
        }

        response = JsonResponse(data)
        assert '"message": "Hello world!"' in response.get_body()
        assert '"today": "{:s}"'.format(isoformat) in response.get_body()
        assert '"undefined": null' in response.get_body()
        assert '"decimal": "1.0008"' in response.get_body()
        assert '"filestorage": "attachment.txt"' in response.get_body()
        assert '"dict": {"a": 12}' in response.get_body()
        assert '"list": [0, 1, 2, 3, 4]' in response.get_body()
        assert '"str": "A piece of text."' in response.get_body()
        assert '"repr": "{:s}"'.format(repr(self)) in response.get_body()

    def test_html(self):
        data = '<p>Hello world!</p>'

        response = HtmlResponse(data)

        assert response.get_body() == '<p>Hello world!</p>'

    def test_template(self):
        data = {
            'message': 'Hello world!'
        }

        response = TemplateResponseTest('index.html', data)

        assert response.get_body() == '<p>Hello world!</p>'

    def test_file(self):
        filename = path.join(HERE, 'environs', 'attachment.txt')
        response = FileResponse(filename)

        assert response.filename == filename
        assert response.as_attachment is False
        assert response.attachment_filename is None
        assert response.size == 19
        assert response.etag.startswith('etag-')
        assert '-19-' in response.etag

        response.headers.clear()
        response.prepare_response()

        wsgi_list = response.headers.to_wsgi_list()
        assert wsgi_list[1:4] == [
            ('Cache-Control', 'public'),
            ('Content-Type', 'text/plain'),
            ('Content-Length', '19')
        ]

        response.headers.clear()
        response.prepare_response(DummyRequest({
            'REQUEST_METHOD': 'GET'
        }))

        wsgi_list = response.headers.to_wsgi_list()
        assert wsgi_list[1:3] == [
            ('Etag', '"' + response.etag + '"'),
            ('Cache-Control', 'max-age=43200, public'),
        ]
        assert wsgi_list[4:-1] == [
            ('Content-Type', 'text/plain'),
            ('Content-Length', '19'),
        ]

        response.headers.clear()
        response.prepare_response(DummyRequest({
            'REQUEST_METHOD': 'GET',
            'HTTP_IF_NONE_MATCH': response.etag,
        }))

        wsgi_list = response.headers.to_wsgi_list()
        assert wsgi_list[1:] == [
            ('Etag', '"' + response.etag + '"'),
            ('Cache-Control', 'max-age=43200, public'),
        ]

        response = FileResponse(filename, as_attachment=True)

        assert response.as_attachment is True
        assert response.attachment_filename == 'attachment.txt'

        response.headers.clear()
        response.prepare_response()

        wsgi_list = response.headers.to_wsgi_list()
        assert wsgi_list[1:4] == [
            ('Cache-Control', 'public'),
            ('Content-Type', 'text/plain'),
            ('Content-Length', '19'),
        ]
        assert wsgi_list[-1] == ('Content-Disposition',
                                 'attachment; filename=attachment.txt')

        response = FileResponse(filename, as_attachment='download.txt')

        assert response.as_attachment is True
        assert response.attachment_filename == 'download.txt'

        with raises(ValueError):
            FileResponse(filename + '-error-please')

    def test_template_or_json(self):
        data = {
            'message': 'Hello world!'
        }

        environ = get_environ('base')
        request = Request(environ)

        response = template_or_json_response(
            request, 'index.html',
            data, template_response_class=TemplateResponseTest)

        assert response

        environ = get_environ('json-post')
        request = Request(environ)

        response = template_or_json_response(
            request, 'index.html',
            data, json_response_class=JsonResponse)

        assert response



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .utils import FileStorage
from decimal import Decimal
from jinja2.runtime import Undefined
from jinja2.exceptions import TemplateNotFound
from os import path
from pandora.compat import PY2
from pandora.compat import text_type
from time import mktime
from time import time
from werkzeug.datastructures import Headers
from werkzeug.http import dump_cookie
from werkzeug.http import http_date
from werkzeug.http import HTTP_STATUS_CODES
from werkzeug.http import is_resource_modified
from werkzeug.wsgi import wrap_file
from zlib import adler32
import datetime as dt
import jinja2
import json
import mimetypes


class Response(object):
    """A response wrapper.

    :param body: Body of the response.
    :param headers: Headers to respond with.
    :param status_code: Status-code of the response [default 200].

    :type body: str
    :type headers: werkzeug.datastructures.Headers | list
    :type status_code: int

    .. code-block:: python

        import poort

        def application(environ, start_response):
            request = poort.Request(environ)
            response = poort.Response('Hallo world!')

            return response(request, start_response)

    """

    def __init__(self, body='', headers=None, status_code=200):
        self.status_code = status_code
        self.headers = Headers(headers)
        self.body = body

    def get_status(self):
        """Get the status of the response as HTTP code.

        :rtype: str

        """

        return '{:d} {:s}'.format(self.status_code,
                                  HTTP_STATUS_CODES[self.status_code])

    def get_body(self):
        """Retrieve the body of the response.

        You can override this method to serialize JSON, for example,
        and return this as the body.

        :rtype: str

        """
        return self.body

    def prepare_response(self, request=None):
        """Prepare the response.

        This prepares (encodes) the body, you can override this method
        to add/override headers when the response is being prepared.

        To yield other content, you should look at `get_body`.

        :param request: Request that was made, None is allowed.

        :type request: poort.Request

        :rtype: str

        """
        if PY2:
            return [self.get_body().encode('utf-8')]
        else:
            return [bytes(self.get_body(), 'UTF8')]

    def respond(self, request):
        """Prepare a tuple to respond with.

        :param request: Request that was made, None is allowed.

        :type request: poort.Request

        :rtype: tuple[str, list[str], str]

        """
        response = self.prepare_response(request)
        return (self.get_status(), self.headers.to_wsgi_list(), response)

    def __call__(self, request, start_response):
        """Respond to WSGI with the current settings.

        :param request: Request that was made, None is allowed.
        :param start_response: Handler provided by WSGI.

        :type request: poort.Request
        :type start_response: callable

        """
        status, headers, response = self.respond(request)
        start_response(status, headers)
        return response

    def set_cookie(self, name, value='', max_age=None,
                   path='/', domain=None, secure=False, httponly=True):
        """Add a cookie to the response.

        :param name: Name of the cookie.
        :param value: Value of the cookie (should always be a string).
        :param max_age: Maximum age (leave `None` for "browser session").
        :param path: Path to bind the cookie to [default `'/'`].
        :param domain: Domain to bind the cookie to [default `None`].
        :param secure: Secure the cookie [default `False`],
            handy: `request.ssl`.
        :param httponly: Cookie not accessable by JavaScript [default `true`].

        :type name: str
        :type value: str
        :type max_age: int | None
        :type path: str
        :type domain: str | None
        :type secure: bool
        :type httponly: bool

        """
        self.headers.add('Set-Cookie', dump_cookie(
            name, value=str(value), max_age=max_age,
            path=path, domain=domain,
            secure=secure, httponly=httponly,
            charset='utf-8', sync_expires=True))

    def del_cookie(self, name, path='/', domain=None, secure=False,
                   httponly=True):
        """Delete a cookie from the browser.

        :param name: Name of the cookie.
        :param path: Path to bind the cookie to [default `'/'`].
        :param domain: Domain to bind the cookie to [default `None`].
        :param secure: Secure the cookie [default `False`],
            handy: `request.ssl`.
        :param httponly: Cookie not accessable by JavaScript [default `true`].

        :type name: str
        :type path: str
        :type domain: str | None
        :type secure: bool
        :type httponly: bool

        .. note::
            Take note that you must match the original settings of the cookie.
            Deleting a cookie with a non-matching path will not work.

        """
        self.set_cookie(name, max_age=0, path=path, domain=domain,
                        secure=secure, httponly=httponly)


class JsonResponse(Response):
    """A response wrapper for JSON results.

    Automaticly sets the *Content-Type* to `application/json` and transforms
    your data when the response is prepared. Override `json_default` to
    provide your own default transformations.

    It automaticly handles `datetime`, `date,`, `Decimal`, `FileStorage`,
    items with `.to_dict()`, items that can be iterated and items with
    `__str__`. When this fails, it simple calls `repr` on the item, it will
    not raise an Exception when it cannot handle formatting the item.

    :param data: Data to convert to JSON.
    :param headers: A list of extra headers, be warned that some headers
        may be overridden.
    :param status_code: Status-Code of the response [default `200`].

    :type data: dict
    :type status_code: int
    :type template_response_class: callable | None

    .. code-block:: python

        import poort

        def get_some_information():
            return poort.JsonResponse({
                'status': 'ok',
                'data': {
                    'some': 'Data here.'
                }
            })

    """

    def __init__(self, data, headers=None, status_code=200):
        self.data = data
        super(JsonResponse, self).__init__(headers=headers,
                                           status_code=status_code)
        self.headers.set('Content-Type', 'application/json')

    def get_body(self):
        return json.dumps(self.data, default=self.json_default)

    def json_default(self, value):
        if type(value) is Undefined:
            return None
        elif isinstance(value, (dt.datetime, dt.date)):
            return value.isoformat()
        elif isinstance(value, Decimal):
            return str(value)
        elif isinstance(value, FileStorage):
            return value.filename
        elif hasattr(value, 'to_dict'):
            return value.to_dict()
        elif hasattr(value, '__iter__'):
            return [v for v in value]
        elif type(value).__str__ is not object.__str__:
            return text_type(value)
        else:
            return repr(value)


class HtmlResponse(Response):
    """A response wrapper for HTML results.

    Automaticly sets the *Content-Type* to `text/html`

    :param body: Your HTML data.
    :param headers: A list of extra headers, be warned that some headers
        may be overridden.
    :param status_code: Status-Code of the response [default `200`].

    :type body: str
    :type status_code: int
    :type template_response_class: callable | None

    .. code-block:: python

        import poort

        def get_some_information():
            return poort.HtmlResponse('''
                <html>
                    <head>
                        <title>Hello</title>
                    </head>
                    <body>
                        <p>Hello, this is some HTML.</p>
                    </body>
                </html>
            ''')

    """

    def __init__(self, body, headers=None, status_code=200):
        super(HtmlResponse, self).__init__(body, headers, status_code)
        self.headers.set('Content-Type', 'text/html')


class TemplateResponse(HtmlResponse):
    """A templated wrapper for HTML results.

    Automaticly sets the *Content-Type* to `text/html`. This uses the
    staticly bound `renderer` (a Jinja2 environment by default) to render
    your HTML.

    :param filename: Filename of the template.
    :param context: Context passed into the template.
    :param headers: A list of extra headers, be warned that some headers
        may be overridden.
    :param status_code: Status-Code of the response [default `200`].

    :type filename: str
    :type context: dict
    :type status_code: int
    :type template_response_class: callable | None

    :ivar jinja2.Environment renderer: Template renderer.

    .. code-block:: python

        import poort

        poort.TemplateResponse.configure('./templates')

        def get_some_information():
            return poort.TemplateResponse('index.html', {'name': 'World'})

    .. code-block:: html

        <!-- templates/index.html -->
        <html>
            <head>
                <title>Hallo</title>
            </head>
            <body>
                <p>Hallo {{ name }}!</p>
            </body>
        </html>

    """

    renderer = None

    def __init__(self, filename, context, headers=None, status_code=200):
        self.filename = filename
        self.context = context

        super(HtmlResponse, self).__init__(headers=headers,
                                           status_code=status_code)

    def get_template(self, filename=None):
        """Get a template from the renderer.

        :param filename: Get the template of `filename`
            [default `self.filename`].

        :type filename: str

        :rtype: jinja2.Template
        """
        return self.renderer.get_template(filename or self.filename)

    def render(self, context=None, filename=None):
        """Render a template.

        :param context: Context passed into the template.
        :param filename: Filename of the template [default `self.filename`].

        :type context: dict
        :type filename: str

        :rtype: str
        """
        if context is None:
            context = dict()

        return self.get_template(filename).render(**context)

    def get_body(self):
        return self.render(context=self.context)

    @classmethod
    def configure(cls, file_system_load_path='./templates', extensions=None):
        """Configure the class.

        This will instantiate the `renderer` variable with a Jinja2
        Environment.

        :param file_system_load_path: Where the templates are located.
        :param extensions: A list of Jinja2 extensions
            [defaults `do, loopcontrols, with_, autoescape`].

        :type file_system_load_path: str
        :type extensions: list[str]

        :rtype: jinja2.Environment
        """
        file_system_load_path = path.abspath(file_system_load_path)
        loader = jinja2.FileSystemLoader(file_system_load_path)

        extensions = extensions or [
            'jinja2.ext.do',
            'jinja2.ext.loopcontrols',
            'jinja2.ext.with_',
            'jinja2.ext.autoescape',
        ]

        cls.renderer = jinja2.Environment(loader=loader, extensions=extensions)
        return cls.renderer


class WrappedTemplateResponse(TemplateResponse):
    """A wrapping templated wrapper for HTML results.

    Automaticly sets the *Content-Type* to `text/html`. You can use this class
    to have a wrapper around your template. It extends `TemplateResponse` and
    works in the same manner, except it also renders a `base_template` and
    injects the rendered template into it as the `body`.

    :param filename: Filename of the template.
    :param context: Context passed into the template.
    :param headers: A list of extra headers, be warned that some headers
        may be overridden.
    :param status_code: Status-Code of the response [default `200`].
    :param base_template:

    :type filename: str
    :type as_attachment: bool | str
    :type cache_timeout: int
    :type headers: werkzeug.datastructures.Headers | list
    :type status_code: int

    .. code-block:: python

        import poort

        poort.WrappedTemplateResponse.configure('./templates')
        poort.WrappedTemplateResponse.BASE_TEMPLATE = 'base.html'

        def get_some_information():
            return poort.TemplateResponse('index.html', {
                'name': 'World',
                'scripts': ['/assets/scripts/jquery.min.js',
                            '/assets/scripts/bootstrap.min.js'],
                'styles': ['/assets/styles/bootstrap.min.css'],
            })

    .. code-block:: html

        <!-- templates/index.html -->
        {% do scripts.append('/assets/scripts/index.js') %}
        <p>Hallo {{ name }}!</p>

    .. code-block:: html

        <!-- templates/base.html -->
        <html>
            <head>
                <title>Hallo</title>

                {% for style in context.styles %}
                  <link href="{{ style }}" rel="stylesheet">
                {% endfor %}
            </head>
            <body>
                {{ body }}
                {% for script in context.scripts %}
                    <script src="{{ script }}">
                {% endfor %}
            </body>
        </html>

    .. note::

        This is a work-around for mutating *scripts* and *styles* (among
        other things) while rendering a page. By mutating the original
        context, the base template can use the *updated* variables. If you
        simply want to render templates without this functionality, simply use
        `TemplateResponse` as the simple extending result can be achieved
        using `{% extend 'base.html' %}`.

    """

    BASE_TEMPLATE = 'base.html'

    def __init__(self, template, context, headers=None, status_code=200,
                 base_template=None):
        self.base_template = base_template or self.BASE_TEMPLATE

        super(WrappedTemplateResponse, self).__init__(template, context,
                                                      headers, status_code)

    def get_body(self):
        return self.render(context={
            'body': self.render(context=self.context, filename=self.filename),
            'context': self.context,
        }, filename=self.base_template)


class FileResponse(Response):
    """A file wrapper response.

    This will send a file to the requester. The *Content-Type* will be
    guessed, it will also set the *Content-Length* and *Last-Modified*
    information based on the file.

    You can enable *caching* by providing the `request` when delivering the
    response and by setting the `cache_timeout`.

    :param filename: Filename you want to sent, absolute is preferred.
    :param as_attachment: Send the file as an attachment, `True` will
        use the basename of `filename`, set it to a `str`
        (`'download.csv'`) to determine the attachment name yourself.
    :param cache_timeout: How long should the result be cached
        [default `43200`, 12 hours).
    :param headers: A list of extra headers, be warned that some headers
        may be overridden.
    :param status_code: Status-Code of the response [default `200`].

    :type filename: str
    :type as_attachment: bool | str
    :type cache_timeout: int
    :type headers: werkzeug.datastructures.Headers | list
    :type status_code: int

    :ivar str attachment_filename: Name of the attachment.

    .. code-block:: python

        import poort

        def get_some_information():
            return poort.FileResponse('cached-123456789-download.csv',
                                      'download-name-here.csv')

        def application(environ, start_response):
            request = poort.Request(environ)
            response = get_some_information()

            return response(request, start_response)

    .. seealso::

        For passing the `request` variable: :py:meth:`Response.__call__`

    """

    def __init__(self, filename, as_attachment=False,
                 cache_timeout=60 * 60 * 12, headers=None, status_code=200):
        if not path.exists(filename):
            raise ValueError('File `{:s}` does not exist.'.format(
                filename))

        super(FileResponse, self).__init__(headers=headers,
                                           status_code=status_code)
        self.filename = filename
        self.cache_timeout = cache_timeout
        self.as_attachment = bool(as_attachment)

        if as_attachment is True:
            self.attachment_filename = path.basename(self.filename)
        elif as_attachment:
            self.attachment_filename = as_attachment
        else:
            self.attachment_filename = None

    @property
    def size(self):
        """Size of the file in bytes.

        :rtype: int
        """
        return path.getsize(self.filename)

    @property
    def mtime(self):
        """Modified datetime of the file.

        :rtype: datetime.datetime
        """
        return dt.datetime.fromtimestamp(int(path.getmtime(self.filename)))

    @property
    def etag(self):
        """Get an Etag based on the file.

        :rtype: str
        """
        if PY2:
            filename = self.filename
        else:
            filename = bytes(self.filename, 'UTF8')

        return 'etag-%d-%s-%s' % (
            mktime(self.mtime.timetuple()),
            self.size,
            adler32(filename) & 0xffffffff
        )

    def prepare_response(self, request=None):
        if request is None:
            environ = dict()
        else:
            environ = request.environ

        guessed_type = mimetypes.guess_type(self.filename)
        mime_type = guessed_type[0] or 'text/plain'

        stream = open(self.filename)

        self.headers['Date'] = http_date()

        if self.cache_timeout and request:
            self.headers.extend((
                ('Etag', '"{:s}"'.format(self.etag)),
                ('Cache-Control', 'max-age={:d}, public'.format(
                    self.cache_timeout)),
            ))

            if not is_resource_modified(environ, self.etag):
                stream.close()
                self.status_code = 304
                return []

            self.headers['Expires'] = http_date(time() + self.cache_timeout)
        else:
            self.headers['Cache-Control'] = 'public'

        self.headers.extend((
            ('Content-Type', mime_type),
            ('Content-Length', str(self.size)),
            ('Last-Modified', http_date(self.mtime))
        ))

        if self.as_attachment:
            self.headers.set('Content-Disposition', 'attachment',
                             filename=self.attachment_filename)

        return wrap_file(environ, stream)


class TerminatingResponse(Exception):
    """Terminating response, when your API does not want to continue.

    Using this exception you can enforce a halt of execution in your
    application without testing your api results.

    :param response: The response you want to pass with this exception.

    :type response: poort.Response

    .. code-block:: python

        import poort

        def ensure_correct_rights():
            # do something with a session or/and database
            valid = False

            if not valid:
                raise poort.TerminatingResponse(JsonResponse({
                    'message': 'You are not allowed here.'
                }, status_code=403))

        def get_some_information():
            return poort.FileResponse('cached-123456789-download.csv',
                                      'download-name-here.csv')

        def application(environ, start_response):
            request = poort.Request(environ)

            try:
                response = get_some_information()
            except poort.TerminatingResponse as tr:
                response = tr.response

            return response(request, start_response)

    """

    def __init__(self, response):
        self.response = response


def template_or_json_response(request, template, context, headers=None,
                              status_code=200, template_response_class=None,
                              json_response_class=None):
    """Return a JSON or HTML response, based on the request.

    When called, it checks the **Request.want_json** variable.
    Based on this outcome a `template_response_class` or a
    `json_response_class` is spawned.

    :param request: Request of the caller.
    :param template: The filename of the template.
    :param context: Context that should be passed.
    :param headers: A list of extra headers, be warned that some headers
        may be overridden.
    :param status_code: Status-Code of the response [default `200`].
    :param template_response_class: Which template response class should be
        used [default :class:`TemplateResponse`].
    :param json_response_class: Which JSON response class should be
        used [default :class:`JsonResponse`].

    :type request: poort.Request
    :type template: str
    :type context: dict
    :type headers: werkzeug.datastructures.Headers | list
    :type status_code: int
    :type template_response_class: callable | None
    :type json_response_class: callable | None

    .. code-block:: python

        import poort
        import poort.template_or_json_response as toj

        def get_some_information(request):
            return toj(request, 'user.html', {
                'id': 44,
                'email': 'j.doe@example.com',
                'first': 'John',
                'last': 'Doe',
            })

        def application(environ, start_response):
            request = poort.Request(environ)
            response = get_some_information(request)

            return response(request, start_response)

    """

    if template_response_class is None:
        template_response_class = TemplateResponse
    if json_response_class is None:
        json_response_class = JsonResponse

    trc = template_response_class
    jrc = json_response_class

    if request.want_json:
        return jrc(context, headers, status_code)
    else:
        try:
            return trc(template, context, headers, status_code)
        except TemplateNotFound:
            return jrc(context, headers, status_code)


def wrapped_template_or_json_response(request, template, context, headers=None,
                                      status_code=200,
                                      template_response_class=None,
                                      json_response_class=None):
    """Return a JSON or HTML response, based on the request.

    When called, it checks the **Request.want_json** variable.
    Based on this outcome a `template_response_class` or a
    `json_response_class` is spawned.

    :param request: Request of the caller.
    :param template: The filename of the template.
    :param context: Context that should be passed.
    :param headers: A list of extra headers, be warned that some headers
        may be overridden.
    :param status_code: Status-Code of the response [default `200`].
    :param template_response_class: Which template response class should be
        used [default :class:`WrappedTemplateResponse`].
    :param json_response_class: Which JSON response class should be
        used [default :class:`JsonResponse`].

    :type request: poort.Request
    :type template: str
    :type context: dict
    :type headers: werkzeug.datastructures.Headers | list
    :type status_code: int
    :type template_response_class: callable | None
    :type json_response_class: callable | None

    .. code-block:: python

        import poort
        import poort.wrapper_template_or_json_response as wtoj

        def get_some_information(request):
            return wtoj(request, 'user.html', {
                'id': 44,
                'email': 'j.doe@example.com',
                'first': 'John',
                'last': 'Doe',
            })

        def application(environ, start_response):
            request = poort.Request(environ)
            response = get_some_information(request)

            return response(request, start_response)

    """

    if template_response_class is None:
        template_response_class = WrappedTemplateResponse
    if json_response_class is None:
        json_response_class = JsonResponse

    return template_or_json_response(request, template, context, headers,
                                     status_code, template_response_class,
                                     json_response_class)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .utils import FileStorage
from .utils import flatten_multidict
from werkzeug.formparser import parse_form_data
from werkzeug.http import parse_cookie
from werkzeug.urls import url_decode
import json


class Request(object):
    """A request wrapper.

    :param host: Hostname of the request.
    :param path: Request part of the URI.
    :param method: Request method (like GET, POST, PUT).
    :param ssl: Check if the request was made over HTTPS.
    :param ip_address: IPAddress of the remote user.
    :param accept_type: Accept header (eg. application/json).
    :param content_type: Content-Type header (eg. multipart/form-data).
    :param query: Query part of the URI (?foo=bar).
    :param form: Form part of the Request (foo=bar).
    :param files: Files part of the Request, removed from `form`.
    :param cookies: Cookies of the Request.
    :param json: JSON part of the Request (content_type is application/json).
    :param params: All params (query, form, files, cookies, json) combined.
    :param environ: The original environ.

    :type host: str
    :type path: str
    :type method: str
    :type ssl: bool
    :type ip_address: str | None
    :type accept_type: str
    :type content_type: str
    :type want_json: bool
    :type want_html: bool
    :type query: dict
    :type form: dict
    :type files: dict[str, poort.utils.FileStorage]
    :type cookies: dict
    :type json: dict | None
    :type params: dict
    :type environ: dict

    .. code-block:: python

        import poort

        def application(environ, start_response):
            request = poort.Request(environ)

            if request.path == '/':
                response = poort.Response('Hallo world!')
            else:
                response = poort.Response('Not found.', status_code=404)

            return response(request, start_response)

    """
    def __init__(self, environ):
        self.environ = environ

        self.host = environ.get('HTTP_HOST', 'localhost')
        self.path = environ.get('PATH_INFO', '/')
        self.method = environ.get('REQUEST_METHOD', 'GET')
        self.ssl = environ.get('HTTPS', False) is not False

        self.ip_address = environ.get('REMOTE_ADDR', None)

        self.accept_type = environ.get('HTTP_ACCEPT', '')
        self.content_type = environ.get('CONTENT_TYPE', '')

        self.want_json = 'application/json' in self.accept_type
        self.want_html = 'text/html' in self.accept_type

        if 'wsgi.input' in environ:
            stream, form, files = parse_form_data(environ)

            self.query = flatten_multidict(url_decode(
                environ.get('QUERY_STRING', '')))
            self.form = flatten_multidict(form)
            self.files = flatten_multidict(files, lambda v: len(v.filename),
                                           FileStorage.from_original)
        else:
            stream = None

            self.query = dict()
            self.form = dict()
            self.files = dict()

        self.cookies = parse_cookie(environ)

        if 'application/json' in self.content_type and stream:
            try:
                self.json = json.loads(stream.read().decode('utf-8'))
            except ValueError:
                self.json = None
        else:
            self.json = None

        self.params = dict()
        self.params.update(self.query)
        self.params.update(self.form)
        self.params.update(self.files)
        if self.json:
            self.params.update(self.json)

    def as_dict(self):
        """Return the full request as a dict

        :rtype: dict
        """
        return {
            'host': self.host,
            'path': self.path,
            'method': self.method,
            'ssl': self.ssl,

            'ip_address': self.ip_address,

            'accept_type': self.accept_type,
            'content_type': self.content_type,

            'want_json': self.want_json,
            'want_html': self.want_html,

            'query': self.query,
            'form': self.form,
            'files': self.files,
            'cookies': self.cookies,
            'json': self.json,

            'params': self.params,
        }

    def get_cookie(self, name, default=None):
        """Retrieve cookie from the request.

        This matches with the api of `poort.Response`, this provides
        `response.set_cookie` and `response.del_cookie`.

        :param name: Name of the cookie.
        :param default: Default when the cookie is not set.

        :type name: str
        :type default: str | None

        :rtype: str | None
        """
        return self.cookies.get(name, default)

    def __repr__(self):
        return '%s %s' % (self.method, self.path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from os import path
from os import SEEK_END
from os import system
from os import unlink
from psutil import Process
import click
import curses
import importlib
import psutil
import sys

try:
    import gunicorn

    assert gunicorn
except ImportError:
    raise ValueError('Poort::cli requires Gunicorn to run!')

_config = None


def config():
    global _config

    if _config is None:
        try:
            sys.path.append('.')
            _config = importlib.import_module('config')
        except ImportError:
            raise ValueError('Could not import `config.py`.')

    return _config


@click.group()
@click.option('--debug/--no-debug', default=False)
def main(debug):
    if debug:
        click.echo('proc_name = %s' % config().proc_name)
        click.echo('bind      = %s' % config().bind)
        click.echo('pidfile   = %s' % config().pidfile)


def get_process():
    if not path.exists(config().pidfile):
        return False

    try:
        with open(config().pidfile) as stream:
            return Process(int(stream.read().strip()))
    except psutil.NoSuchProcess:
        unlink(config().pidfile)
        return False


@main.command('start')
@click.option('--package', default='app')
@click.option('--runnable', default='application')
@click.option('--environ', default='development')
@click.option('--check/--no-check', default=True)
@click.option('--reload', is_flag=True, default=False)
def start(package, runnable, environ, check, reload):
    """Start a server for your app.

    :param package: Package/module containing your app.
    :param runnable: Entrypoint of the server for requests.
    :param environ: Which environment to start
        (production, staging, development [default]).
    :param check: Check if the package is importable and the entrypoint is
        runnable.

    :type package: str
    :type runnable: str
    :type environ: str
    :type check: bool

    .. code-block:: bash

        poort start --environ production
        poort status

    """
    if get_process() is not False:
        click.secho('Application is already running.', err=True, fg='red')
        if reload:
            click.secho('Restarting application.', fg='green')
            system('kill -HUP `cat %s`' % config().pidfile)
        else:
            click.echo('Provide --reload to reload when already started.')
            raise click.Abort

    if check:
        cmd = '%s -c \'import %s; exit(0 if hasattr(%s, "%s") else 1)\''
        cmd = cmd % (sys.executable, package, package, runnable)
        msg = 'Failed to import %s:%s.' % (package, runnable)
        assert system(cmd) == 0, msg

    click.secho('Starting your application.', fg='green')
    system('ENVIRON=%s gunicorn -c config.py %s:%s' % (
        environ, package, runnable))


@main.command('stop')
@click.option('--graceful/--quick', default=True)
def stop(graceful):
    """Stop the server.

    :param graceful: Graceful or forceful (--quick).

    :type graceful: bool

    .. code-block:: bash

        poort stop

    """
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    if graceful:
        click.secho('Stopping your application.', fg='green')
        system('kill -TERM `cat %s`' % config().pidfile)
    else:
        click.secho('Stopping your application (force-quit).', fg='purple')
        system('kill -QUIT `cat %s`' % config().pidfile)


@main.command('reload')
def reload():
    """Reload the server.

    .. code-block:: bash

        poort reload

    """
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    click.secho('Restarting application.', fg='green')
    system('kill -HUP `cat %s`' % config().pidfile)


@main.command('scale')
@click.argument('way', type=click.Choice(['up', 'down']))
@click.argument('amount', default=1)
def scale(way, amount):
    """Scale the workers of server up or down.

    :param way: Which way to scale (--way up | --way down)
    :param amount: The amount of workers to scale.

    :type way: str
    :type amount: int

    .. code-block:: bash

        poort scale --way up --amount 2

    """
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    if amount == 0:
        click.secho('Cannot scale 0.', err=True, fg='red')
        raise click.Abort

    if way == 'down':
        click.secho('Scaling application %d down.' % amount, fg='green')
        for i in range(amount):
            system('kill -TTOU `cat %s`' % config().pidfile)
    elif way == 'up':
        click.secho('Scaling application %d up.' % amount, fg='green')
        for i in range(amount):
            system('kill -TTIN `cat %s`' % config().pidfile)


@main.command('status')
@click.option('--delay', default=1)
def status(delay):
    """Show a status screen (refreshing) of the server.

    The status screen shows information about the workers and gives you
    some shortcut keys to handle the server (quit, upscale, download, reload).
    It also shows the last couple of lines from the server error log.

    :param watch: Keep watching
    :param amount: The amount of workers to scale.

    :type way: str
    :type amount: int

    .. code-block:: bash

        poort status

    Output example::

          test-poort-two


          Running with 2 workers (default is 2)

          Name         PID     CPU      Mem
          ----------------------------------------
          Master     27309    0.0%     8.1M
          Worker 1   27316    0.3%    36.8M
          Worker 2   27319    0.3%    36.6M


          Waiting...           -- (q)uit (u)pscale (d)ownscale (r)eload


          [2016-06-30 13:54:13] [26806] [INFO] Worker exiting (pid: 26806)
          [2016-06-30 13:54:13] [26805] [INFO] Worker exiting (pid: 26805)
          [2016-06-30 13:54:13] [26831] [INFO] Booting worker with pid: 26831
          [2016-06-30 13:54:13] [26832] [INFO] Booting worker with pid: 26832

    .. warning::

        This is a very hacky program, do not mash a key!

    .. warning::

        This needs refactoring into some nicer, more maintainable code.

    """
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    process = get_process()
    if not process:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    screen = curses.initscr()
    curses.halfdelay(delay * 10)
    curses.noecho()

    def line(name, proc, y, x=2):
        cpu_percentage = proc.cpu_percent(None)
        memory = proc.memory_info()

        screen.addstr(y, x, '%-10s %5.d   %4.1f%%   %5.1fM' % (
            name, proc.pid,
            cpu_percentage, memory.rss / 1024 / 1024))

    status = 'Waiting...'

    def display(status, running):
        if not process.is_running():
            raise psutil.NoSuchProcess(process.pid)

        children = process.children()
        workers = len(children)

        screen.erase()
        screen.addstr(1, 2, config().proc_name or 'Unnamed')
        screen.addstr(4, 2, 'Running with %d workers (default is %d)' % (
            workers, config().workers))

        screen.addstr(6, 2, '%-10s %5s   %5s   %6s' % (
            'Name', 'PID', 'CPU', 'Mem'))
        screen.addstr(7, 2, '-' * 40)

        line('Master', process, 8)

        for cx, child in enumerate(children):
            line('Worker %d' % (cx + 1), child, 9 + cx)

        usage = '(q)uit (u)pscale (d)ownscale (r)eload'

        screen.addstr(9 + workers + 2, 2, '%-20s -- %s' % (
            status, usage))

        y, x = screen.getmaxyx()
        height = y - 1

        top = 9 + workers + 5
        max_lines = height - top

        with open(config().errorlog) as stream:
            contents = tail(stream, max_lines)

        for lx, content in enumerate(contents):
            screen.addstr(top + lx, 2, content[:x - 4])

        char = screen.getch()
        if char != curses.ERR:
            key = chr(char)

            if key == 'q':
                status = 'Quit'
                running = False
            elif key == 'u':
                status = 'Scaling up'
                system('kill -TTIN `cat %s`' % config().pidfile)
            elif key == 'd':
                status = 'Scaling down'
                system('kill -TTOU `cat %s`' % config().pidfile)
            elif key == 'r':
                status = 'Restarting'
                system('kill -HUP `cat %s`' % config().pidfile)
            else:
                status = 'Unknown key'
        else:
            status = 'Waiting...'

        return status, running

    running = True
    try:
        while running:
            try:
                status, running = display(status, running)
            except KeyboardInterrupt:
                running = False
    except psutil.Error:
        curses.endwin()

        with open(config().errorlog) as stream:
            click.echo(''.join(tail(stream, 24)))

        click.secho('Application has stopped responding.', err=True, fg='red')
    except:
        curses.endwin()
        raise
    else:
        curses.endwin()


def tail(f, lines=1, _buffer=4098):
    """Tail a file and get X lines from the end"""
    # place holder for the lines found
    lines_found = []

    # block counter will be multiplied by buffer
    # to get the block size from the end
    block_counter = -1

    # loop until we find X lines
    while len(lines_found) < lines:
        try:
            f.seek(block_counter * _buffer, SEEK_END)
        except IOError:  # too small or too many lines requested
            f.seek(0)
            lines_found = f.readlines()
            break

        lines_found = f.readlines()

        if len(lines_found) > lines:
            break

        block_counter -= 1

    return lines_found[-lines:]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from avenue import Avenue
from pandora import Laborer
from poort import Gate
from poort import JsonResponse
from poort import Request
from poort import template_or_json_response
from poort.utils import copy_request_environ
import datetime as dt
import poort
import yaml


class TestApp(object):
    def setup_application(self):
        router = Avenue()
        laborer = Laborer()

        class AppGate(Gate):
            def setup(self, environ):
                super(AppGate, self).setup()
                self.attach('request', Request(environ))

            @property
            def request(self):
                """

                :rtype: Request
                """
                return self.retrieve('request')

        app = AppGate()

        @laborer.provider('request')
        def provide_request(kwargs):
            return app.request

        @router.attach(path='/')
        def get_index(request):
            data = {
                'message': 'Hello world!',
                'request': request.as_dict(),
            }

            return template_or_json_response(request, 'index.html', data)

        @router.attach(path='/throws')
        def get_throws(request):
            raise ValueError('Throws errors...')

        @router.attach(path='/static')
        def get_static(request):
            return poort.FileResponse('./index.html')

        @router.attach(path='/download')
        def get_download(request):
            return poort.FileResponse('./index.html',
                                      as_attachment='download-as-index.html')

        @router.attach(path='/set-cookie')
        def get_set_cookie(request):
            value = str(request.params.get('value', 'not-set'))

            response = JsonResponse({
                'message': 'Cookie was set to {:s}.'.format(value)
            })
            response.set_cookie('poort-cookie', value)
            return response

        @router.attach(path='/set-multiple-cookies')
        def get_set_multiple_cookies(request):
            response = JsonResponse({
                'message': 'Setting 10 cookies!'
            })

            for i in range(10):
                response.set_cookie('poort-cookie-{:d}'.format(i), str(i))

            return response

        @router.attach(path='/view-cookie')
        def get_view_cookie(request):
            value = str(request.get_cookie('poort-cookie'))
            response = JsonResponse({
                'message': 'Cookie is currently {:s}.'.format(value)
            })
            return response

        @router.attach(path='/remove-cookie')
        def get_remove_cookie(request):
            response = JsonResponse({
                'message': 'Cookie is removed.'
            })
            response.del_cookie('poort-cookie')
            return response

        def application(environ, start_response):
            copy = copy_request_environ(environ)

            with app(environ):
                request = app.request

                if request.params.get('store') == 'yes':
                    now = dt.datetime.now()
                    filename = './{:s}.yaml'.format(
                        now.strftime('%Y%m%d-%H%M%S'))

                    with open(filename, 'w') as stream:
                        yaml.safe_dump(copy, stream,
                                       allow_unicode=True,
                                       default_flow_style=False,
                                       indent=2)

                path, match = router.match(**request.as_dict())

                if path:
                    try:
                        partial = laborer.provide(path.func)
                        response = partial(**match.kwargs)
                    except Exception as exc:
                        response = JsonResponse({
                            'message': str(exc),
                        }, status_code=500)
                else:
                    response = JsonResponse({
                        'message': 'Page does not exist.'
                    }, status_code=404)

                return response(request, start_response)

        return application

    def test_basics(self):
        application = self.setup_application()
        assert application


application = TestApp().setup_application()

# -- run this test
# gunicorn test_app:application --reload --threads 4



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from os import path
from os import stat
from os import walk
from pandora import absfilename
from stat import S_IXUSR


class TestWebsite(object):
    def test_files_have_correct_header(self):
        filenames = list()

        for root, dirnames, basenames in walk(absfilename('.')):
            if 'src' not in root or 'tests' not in root:
                continue

            for basename in basenames:
                if basename.endswith('.py'):
                    filenames.append(path.join(root, basename))

        filenames.sort()

        allowed_normal = [
            '# -*- coding: utf-8 -*-',
            'from __future__ import absolute_import, division, print_function',
            ''
        ]
        allowed_executable = [
            '#!/usr/bin/env python',
            '# -*- coding: utf-8 -*-',
            'from __future__ import absolute_import, division, print_function',
            ''
        ]

        for filename in filenames:
            with open(filename) as stream:
                lines = [l.rstrip() for l in stream.read().split('\n')][:4]

            if S_IXUSR & stat(filename).st_mode:
                headers = allowed_executable
            else:
                headers = allowed_normal

            msg = '.{:s} does not have the correct header lines.'.format(
                filename[len(absfilename('.')):])
            msg += '\nFOUND:\n'
            msg += '\n'.join(lines)
            msg += '\nEXPECTED:\n'
            msg += '\n'.join(headers)

            assert len(lines) >= len(headers), msg
            for lx, line in enumerate(headers):
                assert line == lines[lx], msg
